import lzma
import pickle
from datetime import date, datetime, timezone
from pathlib import Path
from typing import Sequence, Type, Union

from sqlalchemy.orm import Query, Session, aliased
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm.session import Session
from waveforms.dicttree import foldDict

from .models import (Cell, Comment, Config, InputText, Notebook, Record,
                     Report, Sample, Tag, utcnow)


def tag(session: Session, tag_text: str) -> Tag:
    """Get a tag from the database or create a new if not exists."""
    try:
        return session.query(Tag).filter(Tag.text == tag_text).one()
    except NoResultFound:
        tag = Tag(text=tag_text)
        return tag


def tag_it(session: Session, tag_text: str, obj: Union[Sample, Record,
                                                       Report]) -> Tag:
    """Tag an object."""
    if obj.id is None:
        session.add(obj)
        obj.tags.append(tag(session, tag_text))
    else:
        session.query(type(obj)).filter(
            type(obj).id == obj.id).one().tags.append(tag(session, tag_text))
    session.commit()


def get_object_with_tags(session: Session,
                         cls: Union[Type[Comment], Type[Sample], Type[Record],
                                    Type[Report]], *tags: str) -> Query:
    """
    Query objects with the given tags.

    Parameters
    ----------
    session : :class:`sqlalchemy.orm.Session`
        The database session.
    cls : :class:`sqlalchemy.orm.Mapper`
        The object class.
    tags : str
        The tags.

    Returns
    -------
    :class:`sqlalchemy.orm.Query`
        The query.
    """
    if isinstance(session, Query):
        q = session
    else:
        q = session.query(cls)
    if not hasattr(cls, 'tags'):
        return []

    aliase = {tag: aliased(Tag) for tag in tags}

    for tag, a in aliase.items():
        q = q.join(a, cls.tags)
        if '*' in tag:
            q = q.filter(a.text.like(tag.replace('*', '%')))
        else:
            q = q.filter(a.text == tag)
    return q


def query_record(session: Session,
                 offset: int = 0,
                 limit: int = 10,
                 app: str | None = None,
                 tags: Sequence[str] = (),
                 before: datetime | date | None = None,
                 after: datetime | date | None = None):
    tz_offset = datetime.now(timezone.utc).astimezone().utcoffset()
    table = {'header': ['ID', 'App', 'tags', 'created time'], 'body': []}
    apps = sorted(
        set([
            n for n, *_ in get_object_with_tags(session.query(Record.app),
                                                Record, *tags).all()
        ]))
    apps = foldDict(dict([(app, None) for app in apps]))

    query = get_object_with_tags(session, Record, *tags)

    if app is not None:
        if app.endswith('*'):
            query = query.filter(Record.app.like(app[:-1] + '%'))
        else:
            query = query.filter(Record.app == app)
    if before is not None:
        if isinstance(before, date):
            before = datetime(before.year, before.month, before.day)
        query = query.filter(Record.ctime <= before - tz_offset)
    if after is not None:
        if isinstance(after, date):
            after = datetime(after.year, after.month, after.day)
        query = query.filter(Record.ctime >= after - tz_offset)
    total = query.count()
    for r in query.order_by(Record.ctime.desc()).limit(limit).offset(offset):
        tags = sorted([t.text for t in r.tags])
        ctime = r.ctime + tz_offset
        row = [r.id, r.app, tags, ctime]
        table['body'].append(row)

    return total, apps, table


def update_tags(session: Session,
                record_id: int,
                tags: Sequence[str],
                append: bool = False):
    record = session.get(Record, record_id)
    if record is None:
        return False
    if append:
        old = [t.text for t in record.tags]
        for t in old:
            if t not in tags:
                tags.append(t)
    record.tags = [tag(session, t) for t in tags]
    try:
        session.commit()
    except Exception:
        session.rollback()
        return False
    return True


def remove_tags(session: Session, record_id: int, tags: Sequence[str]):
    record = session.get(Record, record_id)
    if record is None:
        return False
    old = [t.text for t in record.tags]
    record.tags = [tag(session, t) for t in old if t not in tags]
    try:
        session.commit()
    except Exception:
        session.rollback()
        return False
    return True


def create_notebook(session: Session, notebook_name: str) -> Notebook:
    """Create a notebook in the database."""
    notebook = Notebook(name=notebook_name)
    session.add(notebook)
    return notebook


def create_input_text(session: Session, input_text: str) -> InputText:
    """Create an input text in the database."""
    input = InputText()
    input.text = input_text
    try:
        input = session.query(InputText).filter(
            InputText.hash == input.hash,
            InputText.text_field == input_text).one()
    except NoResultFound:
        session.add(input)
    return input


def create_cell(session: Session, notebook: Notebook, input_text: str) -> Cell:
    """Create a cell in the database."""
    cell = Cell()
    cell.notebook = notebook
    cell.input = create_input_text(session, input_text)
    cell.index = len(notebook.cells) - 1
    session.add(cell)
    notebook.atime = cell.ctime
    return cell


def create_config(session: Session, config: dict | bytes, base: Path,
                  filename: str) -> Config:
    """Create a config in the database."""

    if not isinstance(config, bytes):
        buf = pickle.dumps(config)
        buf = lzma.compress(buf)
        content_type = 'application/pickle+lzma'
    else:
        buf = config
        content_type = 'application/octet-stream'
    config = Config(buf)
    config.content_type = content_type
    for cfg in session.query(Config).filter(Config.hash == config.hash).all():
        with open(base / cfg.file, 'rb') as f:
            if f.read() == buf:
                cfg.atime = utcnow()
                return cfg
    else:
        path = base / filename
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, 'wb') as f:
            f.write(buf)
        config.file = filename
        session.add(config)
    return config


def get_config(session: Session, config_id: int, base: Path):
    config = session.get(Config, config_id)
    if config is None:
        return None
    config.atime = utcnow()
    path = base / config.file
    with open(path, 'rb') as f:
        buf = f.read()
    if config.content_type == 'application/pickle+lzma':
        buf = lzma.decompress(buf)
        buf = pickle.loads(buf)
    return buf



from typing import Type

from sqlalchemy import Column, ForeignKey, Integer, String, Table
from sqlalchemy.orm import Query, Session, aliased, relationship
from sqlalchemy.orm.exc import NoResultFound

from . import Base


class Tag(Base):
    __tablename__ = 'tags'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)

    def __init__(self, name: str) -> None:
        self.name = name

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f'Tag({self.name!r})'


def has_tags(cls: Type[Base]) -> Type[Base]:
    table = Table(
        f'{cls.__tablename__}_tags', Base.metadata,
        Column('item_id',
               ForeignKey(f'{cls.__tablename__}.id'),
               primary_key=True),
        Column('tag_id', ForeignKey('tags.id'), primary_key=True))

    cls.tags = relationship("Tag", secondary=table, backref=cls.__tablename__)

    def add_tag(self, tag: Tag):
        if tag not in self.tags:
            self.tags.append(tag)

    def remove_tag(self, tag: Tag):
        if tag in self.tags:
            self.tags.remove(tag)

    cls.add_tag = add_tag
    cls.remove_tag = remove_tag

    return cls


def tag(session: Session, tag_text: str) -> Tag:
    """Get a tag from the database or create a new if not exists."""
    try:
        return session.query(Tag).filter(Tag.text == tag_text).one()
    except NoResultFound:
        tag = Tag(text=tag_text)
        return tag


def get_object_with_tags(session: Session, cls: Type[Base],
                         *tags: str) -> Query:
    """
    Query objects with the given tags.

    Parameters
    ----------
    session : :class:`sqlalchemy.orm.Session`
        The database session.
    cls : :class:`sqlalchemy.orm.Mapper`
        The object class.
    tags : str
        The tags.

    Returns
    -------
    :class:`sqlalchemy.orm.Query`
        The query.
    """
    if isinstance(session, Query):
        q = session
    else:
        q = session.query(cls)
    if not hasattr(cls, 'tags'):
        return []

    aliase = {tag: aliased(Tag) for tag in tags}

    for tag, a in aliase.items():
        q = q.join(a, cls.tags)
        if '*' in tag:
            q = q.filter(a.text.like(tag.replace('*', '%')))
        else:
            q = q.filter(a.text == tag)
    return q



import matplotlib.pyplot as plt
import numpy as np

from ._autoplot import autoplot


def plotLine(c0, c1, ax, **kwargs):
    t = np.linspace(0, 1, 11)
    c = (c1 - c0) * t + c0
    ax.plot(c.real, c.imag, **kwargs)


def plotCircle(c0, r, ax, **kwargs):
    t = np.linspace(0, 1, 1001) * 2 * np.pi
    s = c0 + r * np.exp(1j * t)
    ax.plot(s.real, s.imag, **kwargs)


def plotEllipse(c0, a, b, phi, ax, **kwargs):
    t = np.linspace(0, 1, 1001) * 2 * np.pi
    c = np.exp(1j * t)
    s = c0 + (c.real * a + 1j * c.imag * b) * np.exp(1j * phi)
    ax.plot(s.real, s.imag, **kwargs)


def plotDistribution(s0,
                     s1,
                     fig=None,
                     axes=None,
                     info=None,
                     hotThresh=10000,
                     logy=False):
    from waveforms.math.fit import get_threshold_info, mult_gaussian_pdf

    if info is None:
        info = get_threshold_info(s0, s1)
    else:
        info = get_threshold_info(s0, s1, info['threshold'], info['phi'])
    thr, phi = info['threshold'], info['phi']
    # visibility, p0, p1 = info['visibility']
    # print(
    #     f"thr={thr:.6f}, phi={phi:.6f}, visibility={visibility:.3f}, {p0}, {1-p1}"
    # )

    if axes is not None:
        ax1, ax2 = axes
    else:
        if fig is None:
            fig = plt.figure()
        ax1 = fig.add_subplot(121)
        ax2 = fig.add_subplot(122)

    if (len(s0) + len(s1)) < hotThresh:
        ax1.plot(np.real(s0), np.imag(s0), '.', alpha=0.2)
        ax1.plot(np.real(s1), np.imag(s1), '.', alpha=0.2)
    else:
        _, *bins = np.histogram2d(np.real(np.hstack([s0, s1])),
                                  np.imag(np.hstack([s0, s1])),
                                  bins=50)

        H0, *_ = np.histogram2d(np.real(s0),
                                np.imag(s0),
                                bins=bins,
                                density=True)
        H1, *_ = np.histogram2d(np.real(s1),
                                np.imag(s1),
                                bins=bins,
                                density=True)
        vlim = max(np.max(np.abs(H0)), np.max(np.abs(H1)))

        ax1.imshow(H1.T - H0.T,
                   alpha=(np.fmax(H0.T, H1.T) / vlim).clip(0, 1),
                   interpolation='nearest',
                   origin='lower',
                   cmap='coolwarm',
                   vmin=-vlim,
                   vmax=vlim,
                   extent=(bins[0][0], bins[0][-1], bins[1][0], bins[1][-1]))

    ax1.axis('equal')
    ax1.set_xticks([])
    ax1.set_yticks([])
    for s in ax1.spines.values():
        s.set_visible(False)

    # c0, c1 = info['center']
    # a0, b0, a1, b1 = info['std']
    params = info['params']
    r0, i0, r1, i1 = params[0][0], params[1][0], params[0][1], params[1][1]
    a0, b0, a1, b1 = params[0][2], params[1][2], params[0][3], params[1][3]
    c0 = (r0 + 1j * i0) * np.exp(1j * phi)
    c1 = (r1 + 1j * i1) * np.exp(1j * phi)
    phi0 = phi + params[0][6]
    phi1 = phi + params[1][6]
    plotEllipse(c0, 2 * a0, 2 * b0, phi0, ax1)
    plotEllipse(c1, 2 * a1, 2 * b1, phi1, ax1)

    im0, im1 = info['idle']
    lim = min(im0.min(), im1.min()), max(im0.max(), im1.max())
    t = (np.linspace(lim[0], lim[1], 3) + 1j * thr) * np.exp(-1j * phi)
    ax1.plot(t.imag, t.real, 'k--')

    ax1.plot(np.real(c0), np.imag(c0), 'o', color='C3')
    ax1.plot(np.real(c1), np.imag(c1), 'o', color='C4')

    re0, re1 = info['signal']
    x, a, b, c = info['cdf']

    xrange = (min(re0.min(), re1.min()), max(re0.max(), re1.max()))

    n0, bins0, *_ = ax2.hist(re0, bins=80, range=xrange, alpha=0.5)
    n1, bins1, *_ = ax2.hist(re1, bins=80, range=xrange, alpha=0.5)

    x_range = np.linspace(x.min(), x.max(), 1001)
    *_, cov0, cov1 = info['std']
    ax2.plot(
        x_range,
        np.sum(n0) * (bins0[1] - bins0[0]) *
        mult_gaussian_pdf(x_range, [r0, r1], [
            np.sqrt(cov0[0, 0]), np.sqrt(cov1[0, 0])
        ], [params[0][4], 1 - params[0][4]]))
    ax2.plot(
        x_range,
        np.sum(n1) * (bins1[1] - bins1[0]) *
        mult_gaussian_pdf(x_range, [r0, r1], [
            np.sqrt(cov0[0, 0]), np.sqrt(cov1[0, 0])
        ], [params[0][5], 1 - params[0][5]]))
    ax2.set_ylabel('Count')
    ax2.set_xlabel('Projection Axes')
    if logy:
        ax2.set_yscale('log')
        ax2.set_ylim(0.1, max(np.sum(n0), np.sum(n1)))

    ax3 = ax2.twinx()
    ax3.plot(x, a, '--', lw=1, color='C0')
    ax3.plot(x, b, '--', lw=1, color='C1')
    ax3.plot(x, c, 'k--', alpha=0.5, lw=1)
    ax3.set_ylim(0, 1.1)
    ax3.vlines(thr, 0, 1.1, 'k', alpha=0.5)
    ax3.set_ylabel('Integral Probability')

    return info


ALLXYSeq = [('I', 'I'), ('X', 'X'), ('Y', 'Y'), ('X', 'Y'), ('Y', 'X'),
            ('X/2', 'I'), ('Y/2', 'I'), ('X/2', 'Y/2'), ('Y/2', 'X/2'),
            ('X/2', 'Y'), ('Y/2', 'X'), ('X', 'Y/2'), ('Y', 'X/2'),
            ('X/2', 'X'), ('X', 'X/2'), ('Y/2', 'Y'), ('Y', 'Y/2'), ('X', 'I'),
            ('Y', 'I'), ('X/2', 'X/2'), ('Y/2', 'Y/2')]


def plotALLXY(data, ax=None):
    assert len(data) % len(ALLXYSeq) == 0

    if ax is None:
        ax = plt.gca()

    ax.plot(np.array(data), 'o-')
    repeat = len(data) // len(ALLXYSeq)
    ax.set_xticks(np.arange(len(ALLXYSeq)) * repeat + 0.5 * (repeat - 1))
    ax.set_xticklabels([','.join(seq) for seq in ALLXYSeq], rotation=60)
    ax.grid(which='major')


def plot_mat(rho, title='$\\chi$', cmap='coolwarm'):
    lim = np.abs(rho).max()
    N = rho.shape[0]

    fig = plt.figure(figsize=(6, 4))
    fig.suptitle(title)

    ax1 = plt.subplot(121)
    cax1 = ax1.imshow(rho.real, vmin=-lim, vmax=lim, cmap=cmap)
    ax1.set_title('Re')
    ax1.set_xticks(np.arange(N))
    ax1.set_yticks(np.arange(N))

    ax2 = plt.subplot(122)
    cax2 = ax2.imshow(rho.imag, vmin=-lim, vmax=lim, cmap=cmap)
    ax2.set_title('Im')
    ax2.set_xticks(np.arange(N))
    ax2.set_yticks(np.arange(N))

    plt.subplots_adjust(bottom=0.2, right=0.9, top=0.95)

    cbar_ax = fig.add_axes([0.15, 0.15, 0.7, 0.05])
    cb = fig.colorbar(cax1, cax=cbar_ax, orientation='horizontal')
    plt.show()



from .base import Base
from .tag import Tag, has_tags
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import relationship
from datetime import datetime

@has_tags
class Report(Base):
    __tablename__ = 'reports'

    id = Column(Integer, primary_key=True)
    ctime = Column(DateTime, default=datetime.utcnow)
    mtime = Column(DateTime, default=datetime.utcnow)

    name = Column(String)
    datafile_id = Column(Integer, ForeignKey('files.id'))
    configfile_id = Column(Integer, ForeignKey('files.id'))
    meta_id = Column(Integer, ForeignKey('files.id'))

    datafile = relationship("File", foreign_keys=[datafile_id])
    configfile = relationship("File", foreign_keys=[configfile_id])
    metafile = relationship("File", foreign_keys=[meta_id])



import click

from .monitor.__main__ import main as monitor
from .scan.server import server
from .sys.net.cli import dht
from .visualization.__main__ import plot


@click.group()
def main():
    pass


@main.command()
def hello():
    """Print hello world."""
    click.echo('hello, world')


main.add_command(monitor)
main.add_command(plot)
main.add_command(dht)
main.add_command(server)

if __name__ == '__main__':
    main()



import asyncio
import functools
import inspect
import logging
import platform
import struct
from abc import ABC, abstractmethod
from collections.abc import Awaitable

from ...version import __version__
from .exceptions import RPCError, RPCServerError, RPCTimeout
from .serialize import pack, unpack
from .utils import acceptArg

if platform.system() == "Windows":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

log = logging.getLogger(__name__)  # pylint: disable=invalid-name

msgIDFormat = struct.Struct("!IIQ")

__msgIndex = 1024


def nextMsgID(clientID, sessionID=0):
    global __msgIndex
    __msgIndex += 1
    return msgIDFormat.pack(clientID, sessionID, __msgIndex)


def parseMsgID(msgID):
    """
    return: (clientID, sessionID, msgIndex)
    """
    return msgIDFormat.unpack(msgID)


# message type

RPC_REQUEST = b'\x01'
RPC_RESPONSE = b'\x02'
RPC_PING = b'\x03'
RPC_PONG = b'\x04'
RPC_CANCEL = b'\x05'
RPC_SHUTDOWN = b'\x06'
RPC_CONNECT = b'\x07'
RPC_WELCOME = b'\x08'

RPC_MSGIDSIZE = msgIDFormat.size


class RPCMixin(ABC):
    info = {}

    def start(self):
        pass

    def stop(self):
        pass

    def close(self):
        pass

    @property
    @abstractmethod
    def loop(self):
        """
        Event loop.
        """

    @abstractmethod
    async def sendto(self, data, address):
        """
        Send message to address.
        """

    __rpc_handlers = {
        RPC_CONNECT: 'on_connect',
        RPC_WELCOME: 'on_welcome',
        RPC_PING: 'on_ping',
        RPC_PONG: 'on_pong',
        RPC_REQUEST: 'on_request',
        RPC_RESPONSE: 'on_response',
        RPC_CANCEL: 'on_cancel',
        RPC_SHUTDOWN: 'on_shutdown',
    }

    def parseData(self, data):
        msg_type, msg = data[:1], data[1:]
        if msg_type in [RPC_PING, RPC_PONG, RPC_CONNECT, RPC_WELCOME]:
            return msg_type, msg
        elif msg_type in [RPC_REQUEST, RPC_RESPONSE, RPC_CANCEL, RPC_SHUTDOWN]:
            msgID, msg = msg[:RPC_MSGIDSIZE], msg[RPC_MSGIDSIZE:]
            return msg_type, msgID, msg
        # elif msg_type in [RPC_LONGREQUEST, RPC_LONGRESPONSE]:
        #     msgID, sessionID, msg = msg[:20], msg[20:40], msg[40:]
        #     return msg_type, msgID, sessionID, msg
        else:
            raise RPCError(f'Unkown message type {msg_type}.')

    def handle(self, source, data):
        """
        Handle received data.

        Should be called whenever received data from outside.
        """
        msg_type, *args = self.parseData(data)
        log.debug(f'received request {msg_type} from {source}')
        handler = self.__rpc_handlers.get(msg_type, None)
        if handler is not None:
            getattr(self, handler)(source, *args)
        else:
            log.error(f'No handler found for request {msg_type} from {source}')

    async def pong(self, addr):
        await self.sendto(RPC_PONG, addr)

    def on_ping(self, source, msg):
        log.debug(f"received ping from {source}")
        asyncio.ensure_future(self.pong(source), loop=self.loop)


class RPCClientMixin(RPCMixin):
    _client_defualt_timeout = 10
    __pending = None
    __clientID = 1

    @property
    def pending(self):
        if self.__pending is None:
            self.__pending = {}
        return self.__pending

    @property
    def clientID(self):
        return self.__clientID

    def createPending(self, addr, msgID, timeout=1, cancelRemote=True):
        """
        Create a future for request, wait response before timeout.
        """
        fut = self.loop.create_future()
        self.pending[msgID] = (fut,
                               self.loop.call_later(timeout,
                                                    self.cancelPending, addr,
                                                    msgID, cancelRemote))

        def clean(fut, msgID=msgID):
            if msgID in self.pending:
                del self.pending[msgID]

        fut.add_done_callback(clean)

        return fut

    def cancelPending(self, addr, msgID, cancelRemote):
        """
        Give up when request timeout and try to cancel remote task.
        """
        if msgID in self.pending:
            fut, timeout = self.pending[msgID]
            if cancelRemote:
                self.cancelRemoteTask(addr, msgID)
            if not fut.done():
                fut.set_exception(
                    RPCTimeout(
                        f'Node({self.info}): Wait response from {addr} timeout.'
                    ))

    def cancelRemoteTask(self, addr, msgID):
        """
        Try to cancel remote task.
        """
        asyncio.ensure_future(self.sendto(RPC_CANCEL + msgID, addr),
                              loop=self.loop)

    def close(self):
        self.stop()
        for fut, timeout in list(self.pending.values()):
            fut.cancel()
            timeout.cancel()
        self.pending.clear()

    def setTimeout(self, timeout=10):
        self._client_defualt_timeout = timeout

    def remoteCall(self, addr, methodNane, sessionID=0, args=(), kw={}):
        if 'timeout' in kw:
            timeout = kw['timeout']
        else:
            timeout = self._client_defualt_timeout
        msg = pack((methodNane, args, kw))
        msgID = nextMsgID(self.clientID, sessionID)
        asyncio.ensure_future(self.request(addr, msgID, msg), loop=self.loop)
        return self.createPending(addr, msgID, timeout)

    async def connect(self, addr, authkey=b"", timeout=None):
        if timeout is None:
            timeout = self._client_defualt_timeout
        await self.sendto(RPC_CONNECT + authkey, addr)
        fut = self.createPending(addr, addr, timeout, False)
        msgID = await fut
        clientID, *_ = parseMsgID(msgID)
        if clientID < 1024:
            raise RPCError(f'Connect {addr} fail')
        self.__clientID = clientID

    async def ping(self, addr, timeout=1):
        await self.sendto(RPC_PING, addr)
        fut = self.createPending(addr, addr, timeout, False)
        try:
            return await fut
        except RPCTimeout:
            return False

    async def request(self, address, msgID, msg):
        log.debug(f'send request {address}, {msgID.hex()}, {msg}')
        await self.sendto(RPC_REQUEST + msgID + msg, address)

    async def shutdown(self, address, roleAuth):
        await self.sendto(RPC_SHUTDOWN + nextMsgID(self.clientID) + roleAuth,
                          address)

    def on_welcome(self, source, msg):
        if source in self.pending:
            fut, timeout = self.pending[source]
            timeout.cancel()
            if not fut.done():
                fut.set_result(msg)

    def on_pong(self, source, msg):
        log.debug(f"received pong from {source}")
        if source in self.pending:
            fut, timeout = self.pending[source]
            timeout.cancel()
            if not fut.done():
                fut.set_result(True)

    def on_response(self, source, msgID, msg):
        """
        Client side.
        """
        if msgID not in self.pending:
            return
        fut, timeout = self.pending[msgID]
        timeout.cancel()
        try:
            result = unpack(msg)
        except Exception as e:
            fut.set_exception(e)
            return
        if not fut.done():
            if isinstance(result, Exception):
                fut.set_exception(result)
            else:
                fut.set_result(result)


class RPCServerMixin(RPCMixin):
    __tasks = None
    __sessions = None
    __nextClientID = 1024
    __nextSessionID = 1024

    def info(self):
        return {'version': __version__}

    @property
    def nextClientID(self):
        self.__nextClientID += 1
        return self.__nextClientID

    @property
    def nextSessionID(self):
        self.__nextSessionID += 1
        return self.__nextSessionID

    @property
    def tasks(self):
        if self.__tasks is None:
            self.__tasks = {}
        return self.__tasks

    @property
    def sessions(self):
        if self.__sessions is None:
            self.__sessions = {}
        return self.__sessions

    def createTask(self, msgID, coro, timeout=0):
        """
        Create a new task for msgID.
        """
        if timeout > 0:
            coro = asyncio.wait_for(coro, timeout)
        task = asyncio.ensure_future(coro, loop=self.loop)
        self.tasks[msgID] = task

        def clean(fut, msgID=msgID):
            if msgID in self.tasks:
                del self.tasks[msgID]

        task.add_done_callback(clean)

    def cancelTask(self, msgID):
        """
        Cancel the task for msgID.
        """
        if msgID in self.tasks:
            self.tasks[msgID].cancel()

    def createSession(self, clientID, obj):
        sessionID = self.nextSessionID
        self.sessions[(clientID, sessionID)] = obj
        return sessionID

    def removeSession(self, clientID, sessionID):
        del self.sessions[(clientID, sessionID)]

    def close(self):
        self.stop()
        for task in list(self.tasks.values()):
            task.cancel()
        self.tasks.clear()

    def _unpack_request(self, msg):
        try:
            method, args, kw = unpack(msg)
        except:
            raise RPCError("Could not read packet: %r" % msg)
        return method, args, kw

    @property
    def executor(self):
        return None

    @abstractmethod
    def getRequestHandler(self, methodNane, source, msgID, args=(), kw={}):
        """
        Get suitable handler for request.

        You should implement this method yourself.
        """

    def processResult(self, result, method, msgID):
        return result

    async def handle_request(self, source, msgID, method, args, kw):
        """
        Handle a request from source.
        """
        try:
            func = self.getRequestHandler(method, source=source, msgID=msgID)
            result = await self.callMethod(func, *args, **kw)
            result = self.processResult(result, method, msgID)
        except (RPCError, RPCServerError) as e:
            result = e
        except Exception as e:
            result = RPCServerError.make(e)
        msg = pack(result)
        await self.response(source, msgID, msg)

    async def callMethod(self, func, *args, **kw):
        if 'timeout' in kw and not acceptArg(func, 'timeout'):
            del kw['timeout']
        if inspect.iscoroutinefunction(func):
            result = await func(*args, **kw)
        else:
            result = await self.loop.run_in_executor(
                self.executor, functools.partial(func, *args, **kw))
            if isinstance(result, Awaitable):
                result = await result
        return result

    async def response(self, address, msgID, msg):
        log.debug(f'send response {address}, {msgID.hex()}, {msg}')
        await self.sendto(RPC_RESPONSE + msgID + msg, address)

    def on_connect(self, source, msg):
        log.debug(f"connect from {source}")
        if self.auth(source, msg):
            clientID = self.nextClientID
        else:
            clientID = 0
        msg = pack(self.info())
        asyncio.ensure_future(
            self.sendto(
                RPC_WELCOME + nextMsgID(clientID),  # + msg,
                source),
            loop=self.loop)

    def on_request(self, source, msgID, msg):
        """
        Received a request from source.
        """
        method, args, kw = self._unpack_request(msg)
        self.createTask(msgID,
                        self.handle_request(source, msgID, method, args, kw),
                        timeout=kw.get('timeout', 0))

    def on_shutdown(self, source, msgID, roleAuth):
        if self.is_admin(source, roleAuth):
            raise SystemExit(0)

    def auth(self, source, authkey):
        return True

    def is_admin(self, source, roleAuth):
        return True

    def on_cancel(self, source, msgID, msg):
        self.cancelTask(msgID)
